/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#include "addrdecoder.h"
#include "dsim_trace.h"
#include "system.h"
#include "common.h"

using namespace std;

int main(int argc, char* argv[], char* argp[]) {
  AddrDecoder* decoder;

  int num_channels[] = {1, 2, 4};
  int num_ranks[] = {1, 2, 4};
  int subrank_width[] = {1, 2, 4, 8};

  for (int ch = 0; ch < 3; ch++) {
    for (int rk = 0; rk < 3; rk++) {
      for (int sr_width = 0; sr_width < 4; sr_width++) {
        for (int sig = 1024*subrank_width[sr_width]; sig >= (subrank_width[sr_width]*8); sig = sig / 2)
          {
          printf("ch: %d, rk: %d, sr_width: %d, sig: %d\n", num_channels[ch], num_ranks[rk],
              subrank_width[sr_width], sig);
          decoder = new AddrDecoder
              (num_channels[ch], //UINT32 num_channel,
               num_ranks[rk], //UINT32 num_rank,
               8, //UINT32 num_chip,
               8, //UINT32 num_bank,
               65536, //UINT32 num_row,
               1024, //UINT32 num_column,
               1, //UINT32 chip_width,
               8, //UINT32 burst_length,
               sig*(8/subrank_width[sr_width]), // UINT32 channel_interleaving_granularity
               true, // bool rank_interleaving
               subrank_width[sr_width], //UINT32 subrank_width,
               sig, //UINT32 subrank_interleaving_granularity
               true,   //bool bank_xor_hashing,    
               true,   //bool subrank_xor_hashing,
               true);  //bool channel_xor_hashing)
          decoder->PrintMask(cout);

          DRAMAddr daddr;
//          DRAMAddr::num_subrank = 8;
          ADDRESS paddr = 0;

          // Decode consecutive 64B addresses.
          for (int i=0; i < 16384+3; i++) {
//          for (int i=0; i < 20; i++) {
            // TODO(mjeong): vary size
            decoder->Decode(paddr, 64, &daddr);
            cout << "x" << hex << setw(8) << paddr << setw(0) << "\t" << daddr << endl;
            paddr += 0x40; //TODO(mjeong): to test cw position, increase it by 8
          }

          // TODO(mjeong): also check XORing is working
        }
      }
    }
  }
}
