/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Doe Hyun Yoon
 *          Min Kyu Jeong
 */

#ifndef DSIM_TRACE_H_
#define DSIM_TRACE_H_

#include <fstream>
#include <boost/format.hpp>
#include "common.h"
#include "args.h"

namespace drsim {

using std::endl;

// class for trace sim arguments: superset of Args
class TraceSimArgs : public Args {
  TIME sim_cycles_;
  char* trace_file_name_;
  UINT64 sys_freq_;

public:
  TraceSimArgs() : Args() {
    sim_cycles_ = 0;
    trace_file_name_ = NULL;
    sys_freq_ = 4000;
  }
  void parse_input_args(int argc, char *argv[]);

  GETSET(TIME, sim_cycles);
  GETSET(char*, trace_file_name);
  GETSET(UINT64, sys_freq);

  void print_tracesim_args( ostream &os ) {
    using boost::format;
    os << "[Trace Simulation Configuration]" << endl;
    if (sim_cycles_ == 0) {
      os << "Simulating to the end of the file" << endl;
    } else {
      os << "Simulating to " << sim_cycles_ << " cycles in sys_clk" << endl;
    }
    os << format("%-30s : %s") % "Trace Input: " % trace_file_name_ << endl;
    os << format("%-30s : %d") % "System Clock Freq (MHz)" % sys_freq() << endl;
    os << endl;
    print(os);
  }
};

enum ReqType {
  REQ_READ = 0,
  REQ_WRITE,
  REQ_IFETCH,
  NUM_REQ_TYPE,
}; 

typedef struct event_t {
  ReqType reqType_;
  ADDRESS addrBase_;
  int size_;
  TIME timeCurr_;
  bool scheduled_;
  int priority_;
} TraceReq;

bool read_a_request(TraceReq &rcCurr_req, std::fstream &rfsTrace);

} // namespace drsim

#endif /*DSIM_TRACE_H_*/
