#include <iostream>
#include "base/output.hh"
#include "mem_drsim/drsim.hh"

using std::cout;
using std::endl;

#define MAX_NUM_CORES 1
bool closed_page[MAX_NUM_CORES] = {false};
bool rmode_allow_write_ACTPRE[MAX_NUM_CORES] = {false}; 
/* dsim decoder options */
unsigned channel_interleaving_granularity[MAX_NUM_CORES] = {64};
bool rank_interleaving[MAX_NUM_CORES] = {false};
unsigned subrank_width[MAX_NUM_CORES] = {8};
unsigned subrank_interleaving_granularity[MAX_NUM_CORES] = {64};
bool bank_xor_hashing[MAX_NUM_CORES] = {true};
bool subrank_xor_hashing[MAX_NUM_CORES] = {true};
bool channel_xor_hashing[MAX_NUM_CORES] = {false};
int llc_tag_shift = -1;
bool bank_partition = false;
int num_bank = -1;
int bank_offset = -1;
bool use_ingress_q = false;

DrSim::DrSim(const Params* p)
    : PhysicalMemory(p), numReqSent(0),
      responseDelay(p->response_delay),
      cpuPriorityMod(p->cpu_priority_mod),
      stat_print_period(p->stat_print_period),
      tickEvent(this), statEvent(this), exitEvent(this, p->output_file),
      etEnabled(p->et_enabled), etStartTick(p->et_start), etEndTick(p->et_end),
      etStartStopEvent(this)
{
    // TODO(mjeong): using the old style param passing for now
    char buf[65536];
    strncpy(buf, p->drsim_config.c_str(), 65535);
    char* str1 = strtok(buf, ":");
    char* argv[256];
    int argc = 0;
    argv[argc++] = (char*)"dsim";
    while (1) {
        str1 = strtok(NULL, ":");
        if (str1 == NULL)
            break;
        argv[argc++] = str1;
    }
    args.parse_input_args(argc, argv);

    uint32_t capacity_log2 = 3 + log2(args.num_channel());
    capacity_log2 += log2(args.module_config().num_rank);
    capacity_log2 += log2(args.module_config().num_bank);
    capacity_log2 += log2(args.module_config().num_row);
    capacity_log2 += log2(args.module_config().num_column);

    cout << "DrSim configuration" << endl;
    cout << "capacity_log2         : " << capacity_log2 << endl;
#if 0
    // per-page address map
    if (strcmp(page_addrmap_filename, "") != 0) {
        per_page_map_ = true;
        FILE* ppn_file = fopen(page_addrmap_filename, "r");
        char linebuf[1024];
        if (ppn_file == NULL) perror ("Error opening file");
        while (!feof(ppn_file)) {
            fgets(linebuf, 1024, ppn_file);
            if (linebuf[0] == '#')
                continue;
            char* p = strtok(linebuf, ", ");
            int ppn = atoi(p);
            p = strtok(NULL, ", ");
            int map = atoi(p);
    //        cout << ppn << " => " << map << endl;
            if (map == 8) {
                ppn_srmap[ppn] = 1;
            }
        }
        fclose(ppn_file);
    } else {
        per_page_map_ = false;
    }
#endif

    uint32_t num_channel = args.num_channel();

    int num_threads = 1;
    decoder = new drsim::AddrDecoder*[num_threads];

    // Translating old parameters based on 'num_subrank' and 'subrank_map' to new format.
#if 0
    if (args.num_subrank() != 0 && args.subrank_map() != INVALID_SUBRANK_MAP) {
        for (int32_t core_id = 0; core_id < num_threads; core_id++) {
            subrank_width[core_id] = args.module_config().num_chip / args.num_subrank();
            rank_interleaving[core_id] = args.rank_interleave();
            bank_xor_hashing[core_id] = args.bank_xor_hash();
            subrank_xor_hashing[core_id] = args.subrank_xor_hash();
            channel_xor_hashing[core_id] = args.channel_xor_hash();
            switch (args.subrank_map()) {
                case M2:
                    subrank_interleaving_granularity[core_id] = 64; // assuming 64B cache-line size
                    channel_interleaving_granularity[core_id] = 64 * args.num_subrank();
                    break;

                case M4:
                    subrank_interleaving_granularity[core_id] = 
                        args.module_config().num_column * args.module_config().chip_width * subrank_width[core_id];
                    channel_interleaving_granularity[core_id] = 
                        args.module_config().num_column * args.module_config().chip_width * args.module_config().num_chip;
                    break;

                default:
                    PANIC(false);
            }
        }
    }
#endif
    for (int32_t core_id = 0; core_id < num_threads; core_id++) {
        decoder[core_id] = new drsim::AddrDecoder
                (args.num_channel(),//uint32_t num_channel,
                 args.module_config().num_rank, //uint32_t num_rank,
                 args.module_config().num_chip, //uint32_t num_chip,
                 args.module_config().num_bank, //uint32_t num_bank,
                 args.module_config().num_row,    //uint32_t num_row,
                 args.module_config().num_column, //uint32_t num_column,
                 args.module_config().chip_width, //uint32_t chip_width,
                 args.module_config().burst_length, //uint32_t burst_length,

                 channel_interleaving_granularity[core_id], //uint32_t channel_interleaving_granularity
                 false, // bool bank_interleaving
                 rank_interleaving[core_id],        //bool rank_interleaving
                 subrank_width[core_id],                //uint32_t subrank_width,
                 subrank_interleaving_granularity[core_id],//uint32_t subrank_interleaving_granularity
                 (!bank_partition && bank_xor_hashing[core_id]),         //bool bank_xor_hashing,        
                 llc_tag_shift,                                 //uint32_t bank_xor_hashing_offset,
                 subrank_xor_hashing[core_id],    //bool subrank_xor_hashing,
                 channel_xor_hashing[core_id]); //bool channel_xor_hashing)
    }

    cout << "address decoder format:" << endl;
    printDecoderMask(cout);
    cout << "----------------------------------------------------------------" << endl;

    mc = new drsim::MemoryChannel*[num_channel];
    for (uint32_t i = 0; i < num_channel; i++) {
        mc[i] = new drsim::MemoryChannel(args,
                                   args.module_config(),
                                   args.read_queue_size(),
                                   args.write_queue_size(),
                                   i);
    }

    // initialize clock -- DRAM clock and Sys clock all in MHz
    drsim::clock::xInitialize(args.dram_freq(), args.dram_freq());

    cout << "*** drsim args ***" << endl;
    args.print(cout);
    cout << "----------------------------------------------------------------" << endl;
    drsim::DRAMTiming::print_dram_timing(cout);
    cout << "----------------------------------------------------------------" << endl;

    num_bank = args.module_config().num_bank;
    bank_offset = (int)(decoder[0]->bank_offset());
    use_ingress_q = p->use_ingress_q;

    if (etEnabled) {
        etStartTick = 0;
        etEndTick = -1;

        events = new EventTrace;
        globalTrace = new ET;
        events->setTrace(globalTrace);
        events->init();

        string etFileName = simout.directory() + "/event_trace.et";
        globalTrace->setOutputFile(etFileName.c_str());

        cout << "EVENT TRACING - StartTick: " << etStartTick
             << ", EndTick: " << etEndTick << endl;
    }
}

DrSim::~DrSim()
{
  for (int i = 0; i < args.num_channel(); i++) {
    delete mc[i];
  }
  delete [] mc;
  delete decoder; //TODO(mjeong): leak!
}

bool
DrSim::enqueueable(PacketPtr pkt)
{
    bool is_read = pkt->isRead();
    ADDRESS addr = pkt->getAddr();
    if (inRange2(pkt))
        addr = addr - start2() + (size() - size2());
    int decoderid = 0;
    int channel = decoder[decoderid]->GetChannel(addr);

    if (use_ingress_q)
        return !(mc[channel]->isIngressQueueFull());
    else
        return ((is_read && !mc[channel]->isReadQueueFull()) ||
                (!is_read && !mc[channel]->isWriteQueueFull()));
}

void
DrSim::enqueue(PacketPtr pkt, int port_id)
{
    bool fromGPU = pkt->req->getFlags().isSet(Request::GPU);
    int master_id = fromGPU ? 1 : 0;
    DPRINTF(DrSim, "enq: %d %s 0x%x\n", fromGPU, pkt->cmdString(), pkt->getAddr());

    bool is_read = pkt->isRead();
    ADDRESS addr = pkt->getAddr();
    if (inRange2(pkt)) {
        addr = addr - start2() + (size() - size2());
    }
//    ADDRESS line_addr = (addr >> log_base2(linesize)) << log_base2(linesize);
    ADDRESS line_addr = (addr >> 6) << 6;
    int linesize = pkt->getSize();
    // FIXME(mjeong)
    linesize = 64;

    // code copied from DRAMsim wrapper
    if (pkt->isReadWrite()) {
        fatal("I don't know how to handle ReadWrites!");
    } else if (pkt->isRead() || pkt->isWrite()) {
        readwriteAccesses++;
    } else {
        doAtomicAccess(pkt);
        ports[port_id]->schedSendTiming(pkt, curTick() + 1);
        otherAccesses++;
        return;
    }

    // Decode begin
    DRAMAddr dram_addr;
    uint32_t channel = 0;
    int decoderid = 0;
    channel = decoder[decoderid]->GetChannel(line_addr);
    decoder[decoderid]->Decode(line_addr, linesize, &dram_addr);

    // assuming MC_ENQUEUABLE_HEADER is already called and checked,
    // so MC must be able to accept this req
    //
    Gem5Req* req = new Gem5Req(this,
                               numReqSent++,
                               is_read,
                               line_addr,
                               linesize,
                               dram_addr,
                               master_id,
                               pkt,
                               port_id
                               );
#if 0
    // per-core policy
    assert(core->id < MAX_NUM_CORES);
    req->set_is_closed_page(closed_page[core->id]);
    req->set_allow_rmode_write_ACTPRE(rmode_allow_write_ACTPRE[core->id]);
#endif
    if (!fromGPU) {
        for (int32_t i = 0; i < cpuPriorityMod; i++)
            req->RaisePriority();
    }

    // Do the functional operation now to ensure ordering
    doAtomicAccess(pkt);

    // Ship it!!
    //
    if (use_ingress_q)
        mc[channel]->sendRequestInQ(req);
    else
        mc[channel]->sendRequest(req);

    // ETrace
    if (etEnabled) {
        ET::TransHandle etHandle;
        EventTrace::etTransTypes type = is_read ? events->etRdRequest : events->etWrRequest;
        if (fromGPU)
            type = is_read ? events->etGpuRdRequest : events->etGpuWrRequest;
        string type_str = is_read ? "read" : "write";

        etHandle = addETTrans(globalTrace,
                       events->etTransIDs[type],
                       new ET::Attr(events->etAttrIDs[events->etReqId], req->id()),
                       new ET::Attr(events->etAttrIDs[events->etAddr], req->phy_addr()),
                       new ET::Attr(events->etAttrIDs[events->etReqType], type)
                       );
        req->ETHandle(etHandle);
        addETEvent(etHandle, events->etEventIDs[events->etNewRequest], curTick());
    }
}

void
DrSim::tick()
{
    do {
      drsim::clock::xTick();
      if (drsim::clock::is_dram_ticked()) {
        for (int i = 0; i < args.num_channel(); i++)
          mc[i]->cycle();
      }
    } while (!drsim::clock::is_sys_ticked());

    // TODO(mjeong): tick every cycle no matter if there's work to do
    // DRAMsim ticks only when there's a pending req.
    // Power-down work also needs to be accounted for.
    schedule(tickEvent, curTick() + ticks_per_dram_cycle);
}

void
DrSim::printStat(ostream &os)
{
    os << "*** dsim stats ***" << endl;
    os << "********************************************************" << endl;
    os << "Simulation Finished at " <<
          "master_clk: " << drsim::clock::get_master_clk() <<
          ", sys_clk: "  << drsim::clock::get_sys_clk() <<
          ", dram_clk: " << drsim::clock::get_dram_clk() << endl;
    os << "--------------------------------------------------------" << endl;
    drsim::Stat::printall(os);
    os << "--------------------------------------------------------" << endl;
    drsim::MemoryChannel::printChannelStats(os);
    os << "--------------------------------------------------------" << endl;
}

void
DrSim::periodicStatPrint()
{
    static int period_i = 0;
    string fn = csprintf("drsim_stat_period_%d.txt", period_i++);
    ostream* os = simout.create(fn);
    drsim::Stat::printall_periodic(*os);
    simout.close(os);
    schedule(statEvent, curTick() + stat_print_period);
}

void
DrSim::printStatus()
{
    // NOTE(mjeong):
    // there is no standard way to convert FILE* to ofstream
    // we can write MemoryChannel::printStuatus(FILE* fp), but seems
    // to be an overkill, so just print out to the stderr for now
    for (uint32_t i = 0; i < args.num_channel(); i++) {
        mc[i]->printStatus(cout);
    }
}

void
DrSim::sendResponse(PacketPtr pkt, int port_id)
{
    DPRINTF(DrSim, "DRAM access completed: %s 0x%x\n",
            pkt->cmdString(), pkt->getAddr());

    if (pkt->isResponse())
        ports[port_id]->schedSendTiming(pkt, curTick() + 1 + responseDelay);
    else
        delete pkt;
}

void
DrSim::printDecoderMask(ostream &os)
{
  int num_threads = 1;
  if (decoder != NULL) {
    for (int i = 0; i < num_threads; i++) {
      decoder[i]->PrintMask(os);
    }
  }
}

void
DrSim::wrapUpStats()
{
  for (int i = 0; i < args.num_channel(); i++) {
    mc[i]->wrapUpStats();
  }
}

void
DrSim::dumpStates()
{
  for (int i = 0; i < args.num_channel(); i++) {
    mc[i]->printStatus(std::cerr);
  }
}

void 
DrSim::init()
{
    if (ports.size() == 0)
        fatal("PhysicalMemory object %s is unconnected!", name());

    for (vector<DrSimPort*>::iterator pi = ports.begin();
         pi != ports.end(); ++pi) {
        if (*pi)
            (*pi)->sendStatusChange(Port::RangeChange);
    }
}

Port*
DrSim::getPort(const std::string &if_name, int idx)
{
    if (if_name == "functional")
        return new DrSimPort(name() + "-functional", this, -1);

    if (if_name != "port")
        panic("Octopus::getPort: unknown port '%s' requested", if_name);

    if (idx >= ports.size())
        ports.resize(idx+1);

    if (ports[idx] != NULL)
        panic("DrSim::getPort: port %d already assigned", idx);

    DrSimPort *port =
        new DrSimPort(csprintf("%s-port%d", name(), idx), this, idx);

    ports[idx] = port;

    return port;
}

void
DrSim::regStats()
{
    using namespace Stats;

    totalAccesses
        .name(name() + ".totalAccesses")
        .desc("total accesses sent to DRAMsim")
        ;
    readwriteAccesses
        .name(name() + ".readnwriteAccesses")
        .desc("total number of accesses sent to DRAM module")
        ;
    otherAccesses
        .name(name() + ".otherAccesses")
        .desc("total number of accesses not sent to DRAM module")
        ;
    registerExitCallback(&exitEvent);
}

void
Gem5Req::completed_callback()
{
    // Trans finished
    if (etHandle_.get() != 0) {
        addETEvent(etHandle_,
                   drsim_->events->etEventIDs[drsim_->events->etCompleteReq],
                   curTick());
    }

    drsim_->sendResponse(pkt_, port_id_);
}

void
Gem5Req::critical_word_callback(int cw_size)
{
    // No critical word first for Gem5 yet
//    if (pkt_->isResponse())
//        ports[port_id_]->schedSendTiming(pkt_, curTick() + 1);
}

void
Gem5Req::PrintExtra(ostream &os) const
{
    os << " m5tick:" << curTick();
}

DrSim* DrSimParams::create()
{
    return new DrSim(this);
}

void
DrSim::ExitCallbackEvent::process()
{
    drsim->wrapUpStats();
    ostream* os = simout.create(output_fn);
    drsim->printStat(*os);
    simout.close(os);
    if (drsim->etEnabled)
        drsim->events->globalTrace->forceFlush();
}

void
DrSim::etStartStop()
{
    assert((curTick() == etStartTick) || (curTick() == etEndTick));

    if (curTick() == etStartTick) {
        globalTrace->setEnabled(true);
        if ((etEndTick != -1) && !etStartStopEvent.scheduled())
            schedule(etStartStopEvent, etEndTick);
    } else if ((curTick() >= etStartTick) && (curTick() == etEndTick)) {
       globalTrace->setEnabled(false);
    }
}

DrSim::DrSimPort::DrSimPort(string name, DrSim* d, int i)
    : SimpleTimingPort(name, d), id(i), drsim(d)
{
}

bool
DrSim::DrSimPort::recvTiming(PacketPtr pkt)
{
    if (pkt->memInhibitAsserted()) {
        // C&P from SimpleTimingPort::recvTiming()
        // snooper will supply based on copy of packet
        // still target's responsibility to delete packet
        delete pkt;
        return true;
    }

    if (drsim->enqueueable(pkt)) {
        drsim->enqueue(pkt, id);
        return true;
    }
    return false;
}

void
DrSim::DrSimPort::recvFunctional(PacketPtr pkt)
{
    pkt->pushLabel(drsim->name());

    if (!checkFunctional(pkt)) {
        // Default implementation of SimpleTimingPort::recvFunctional()
        // calls recvAtomic() and throws away the latency; we can save a
        // little here by just not calculating the latency.
        drsim->doFunctionalAccess(pkt);
    }

    pkt->popLabel();
}

void
DrSim::startup()
{
    // FIXME(mjeong): assuming 1ps gem5 tick for now.
    ticks_per_dram_cycle = 1000000 / args.dram_freq();
    schedule(tickEvent, curTick() + ticks_per_dram_cycle);

    if (stat_print_period > 0)
        schedule(statEvent, curTick() + stat_print_period);

    if (etEnabled) {
        if (etStartTick == 0) {
            globalTrace->setEnabled(true);
            if ((etEndTick != -1) && !etStartStopEvent.scheduled())
                schedule(etStartStopEvent, etEndTick);
        }
        else {
            assert((etStartTick > 0) && "etStartTick cannot be less than 0\n" );
            if (!etStartStopEvent.scheduled())
                schedule(etStartStopEvent, etStartTick);
        }
    }
}
