/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef LOG_H_

namespace drsim {

#define MAX_DBG_LEVEL 20
#define TRACE_TM 1        // trace for timing model
#define TRACE_MC 2        // trace for memory channel
#define TRACE_DECODER 3   // address decoder
#define TRACE_SCHEDULER 4  // scheduler
#define TRACE_SCHEDULER_V 5  // scheduler verbose
#define TRACE_REFRESH     // refresh
#define TRACE_PDN 6       // power-down control
#define TRACE_STAT 7      // statistics
#define TRACE_CLOCK 9     // trace for clocking scheme

#ifdef LOG_ENABLE
  extern bool DBG_LEVEL[MAX_DBG_LEVEL];
  extern std::ostream* _dbg_log_out;
  extern char _dbg_log_filename[];
  extern UINT32 _dbg_log_purge;
  extern TIME _dbg_log_begin;
  extern TIME _dbg_log_end;

#define DBG_LOG_NOCLK(xx, ...) \
  if (DBG_LEVEL[xx] \
      && clock::get_dram_clk() >= _dbg_log_begin \
      && clock::get_dram_clk() <= _dbg_log_end) {\
    *_dbg_log_out << __VA_ARGS__;\
  }
#define DBG_LOG_ASIS(xx, ...) \
  if (DBG_LEVEL[xx] \
      && clock::get_dram_clk() >= _dbg_log_begin \
      && clock::get_dram_clk() <= _dbg_log_end) {\
    __VA_ARGS__;\
  }
#define DBG_LOG(xx, ...) \
  if (DBG_LEVEL[xx] \
      && clock::get_dram_clk() >= _dbg_log_begin \
      && clock::get_dram_clk() <= _dbg_log_end) {\
    *_dbg_log_out << "[" << std::setw(8) << std::setfill('0') << clock::get_dram_clk() << "] ";\
    *_dbg_log_out << __VA_ARGS__;\
  }
#else
#define DBG_LOG_NOCLK(...) {}
#define DBG_LOG_ASIS(...) {}
#define DBG_LOG(...) {}
#endif

}  // namespace drsim

#endif
