/*
 * Copyright (c) 2012, The University of Texas at Austin.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holder nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Min Kyu Jeong
 */

#ifndef PDNCONTROLLER_H_
#define PDNCONTROLLER_H_

#include "timing_model.h"
#include "stat.h"

namespace drsim {

class PDNController {
  public:
  PDNController(UINT32 channel_id,
      UINT32 num_rank,
      UINT32 num_chip,
      ReqBuffer* read_queue,
      ReqBuffer* write_queue,
      InFlightReqBuffer* inflight_read_queue,
      cTimingModel* timing_logic);
  ~PDNController();
  // upon sim completion, call this to wrap up open stat counters
  void wrapUpStats();

  bool findPDNCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd);
  bool findPUPCandidate(DRAMAddr* daddr, eMemCmd* dram_cmd, Req** cand, bool readonly);

  void activated(const DRAMAddr& daddr);
  void precharged(const DRAMAddr& daddr);
  void powerdown(const DRAMAddr& daddr);
  void powerup(const DRAMAddr& daddr);

  bool is_any_sr_powerdown(const DRAMAddr& daddr);
  bool is_any_sr_powerdown(const DRAMAddr& daddr, ChipMask* pdn_chips);

  ostream& printPDNState(ostream &os);
  UINT64 get_num_cycles_act_pdn_in_sr(UINT32 rk, UINT32 sr);
  UINT64 get_num_cycles_pre_pdn_in_sr(UINT32 rk, UINT32 sr);

  private:
  void updatePowerStatusInQueues();

  UINT32 channel_id_;
  UINT32 num_rank_;
  UINT32 num_chip_;

  // ptr to the global structures
  ReqBuffer* read_queue_;
  ReqBuffer* write_queue_;
  InFlightReqBuffer* inflight_read_queue_;
  cTimingModel* timing_logic_;

  PowerState** chip_pdn_state_;  // [rank][chip]

  // stats
  TIME** last_pre_pdn_in_sr_;  // [rank][chip]
  TIME** last_act_pdn_in_sr_;
  ArrayCounterStat num_cycles_pre_pdn_in_sr_;  // [rk*num_sr + sr]
  ArrayCounterStat num_cycles_act_pdn_in_sr_;  // [rk*num_sr + sr]
};

}  // namespace drsim

#endif
