#!/usr/bin/perl
# Perform a series of regression tests.
# Runs simulation with diffrent configurations on different access-pattern
#  traces. Diff the result with the reference results.
#
# Author: Min Kyu Jeong

use Getopt::Long;
require "dsim_run_helper.pl";

%easyname = ("M2", "CI", "M4", "PI");
my $num_ABUS = 2;
my $generate_ref = 0;
GetOptions("generate_ref" => \$generate_ref);

# stride
system("rm -f result_stride");
print_header("result_stride");
print "stride: running 56 configurations\n";
foreach my $subrank_map ("M2", "M4") {
  foreach my $subrank (1, 2, 4, 8) {
    foreach my $stride (1, 2, 4, 8, 16, 32, 64, 128, 256) {
      my $sys_clk = run("64x${stride}rd_x1_4096.64B.trace", # tracename
                        $subrank,                           # subrank
                        1,                                  # rank
                        32*$subrank,                        # rq_size 
                        32,                                 # wq_size
                        $subrank_map,
                        ($subrank < $num_ABUS) ? $subrank : $num_ABUS,
                        "result_stride");
      if ($sys_clk == -1) {
        die "RUN FAILED: 64x${stride}rd_x1_4096.64B.trace, $easyname{$subrank_map}, num_sr: $subrank\n";
      }
      else {
        print ".";
      }
    }
  }
}
print "\n";
if ($generate_ref) {
  system("mv result_stride result_stride.reference");
}
else {
  print "stride: diff with reference\n";
  system("diff result_stride result_stride.reference");
  print "stride: end of diff with reference\n";
}

## random-consecutive
system("rm -f result_random_consec");
print_header("result_random_consec");
print "random-consecutive: 56 configurations\n";
foreach my $subrank_map ("M2", "M4") {
  foreach my $subrank (1, 2, 4, 8) {
    foreach my $consec (1, 2, 4, 8, 16, 32, 64, 128, 256) {
      my $sys_clk = run("r64+${consec}rd_x1_4096.64B.trace", # tracename
                        $subrank,                           # subrank
                        1,                                  # rank
                        32*$subrank,                        # rq_size 
                        32,                                 # wq_size
                        $subrank_map,
                        $subrank,
                        "result_random_consec");
      if ($sys_clk == -1) {
        die "RUN FAILED: r64+${consec}rd_x1_4096.64B.trace, $easyname{$subrank_map}, num_sr: $subrank\n";
      }
      else {
        print ".";
      }
    }
  }
}
print "\n";
if ($generate_ref) {
  system("mv result_random_consec result_random_consec.reference");
}
else {
  print "random-consecutive: diff with reference\n";
  system("diff result_random_consec result_random_consec.reference");
  print "random-consecutive: end of diff with reference\n";
}

## stride-consecutive
system("rm -f result_stride_consec");
print_header("result_stride_consec");
print "stride-consecutive: 56 configurations\n";
foreach my $subrank_map ("M2", "M4") {
  foreach my $subrank (1, 2, 4, 8) {
    foreach my $consec (1, 2, 4, 8, 16, 32, 64, 128, 256, 512) {
      my $sys_clk = run("64x512+${consec}rd_x1_4096.64B.trace", # tracename
                        $subrank,                           # subrank
                        1,                                  # rank
                        32*$subrank,                        # rq_size 
                        32,                                 # wq_size
                        $subrank_map,
                        $subrank,
                        "result_stride_consec");
      if ($sys_clk == -1) {
        die "RUN FAILED: 64x512+${consec}rd_x1_4096.64B.trace, $easyname{$subrank_map}, num_sr: $subrank\n";
      }
      else {
        print ".";
      }
    }
  }
}
print "\n";
if ($generate_ref) {
  system("mv result_stride_consec result_stride_consec.reference");
}
else {
  print "stride-consecutive: diff with reference\n";
  system("diff result_stride_consec result_stride_consec.reference");
  print "stride-consecutive: end of diff with reference\n";
}
