#!/usr/bin/perl
use POSIX;
use Getopt::Long;

($#ARGV == 4) or ($#ARGV == 3) or print_usage();

# positional arguments
$format = $ARGV[0];
$num_stream = $ARGV[1];
$stream_length = $ARGV[2];
$access_granularity = $ARGV[3];
if ($#ARGV == 4) {
  $trace_file_name = $ARGV[4];
}
else {
  $trace_file_name = "${format}_x${num_stream}_${stream_length}.${access_granularity}B.trace";
}

$time = 0;
$ag_width = 1; 
$injection_interval = 1;
# NOTE(mjeong): getopt doesn't play nice with ARGV == 4 check above.
#               hardcode it for now
#my $result = GetOptions('ag_width=i' => \$ag_width);

@stream_rw = ( "READ", "READ" );
$stream_type = "stride";
$index_range = 4*1024*1024;

if( $format =~ m/^\d+x\d+rd$/ ) {
	$format =~ /^?(\d+)/;
	$element_size = $1;
	$format =~ /x?(\d+)rd$/;
	$stride = $1;
} elsif( $format =~ m/^\d+x\d+rw$/ ) {
	$format =~ /^?(\d+)/;
	$element_size = $1;
	$format =~ /x?(\d+)rw$/;
	$stride = $1;
	$stream_rw[1] = "WRITE";
} elsif( $format =~ m/^(\d+)x(\d+)\+(\d+)rd$/ ) {
	$element_size = $1;
	$stride = $2;
  $consec_num = $3;
  $stream_type = "stride_consec";
} elsif( $format =~ m/^(\d+)x(\d+)\+(\d+)rw$/ ) {
	$element_size = $1;
	$stride = $2;
  $consec_num = $3;
	$stream_rw[1] = "WRITE";
  $stream_type = "stride_consec";
} elsif( $format =~ m/^r\d+rd$/ ) {
	$format =~ /^r?(\d+)rd$/;
	$element_size = $1;
	$stream_type = "indexed";
} elsif( $format =~ m/^r\d+rw$/ ) {
	$format =~ /^r?(\d+)rw$/;
	$element_size = $1;
	$stream_rw[1] = "WRITE";
	$stream_type = "indexed";
} elsif( $format =~ m/^r\d+\+\d+rd$/ ) {
  $format =~ /^r(\d+)\+(\d+)rd$/;
	$element_size = $1;
	$consec_num = $2;
	$stream_type = "indexed_consec";
} elsif( $format =~ m/^r\d+\+\d+rw$/ ) {
  $format =~ /^r(\d+)\+(\d+)rw$/;
	$element_size = $1;
	$consec_num = $2;
	$stream_rw[1] = "WRITE";
	$stream_type = "indexed_consec";
} elsif( $format =~ m/^cr\d+rd$/ ) {
	$format =~ /^cr?(\d+)rd$/;
	$element_size = $1;
	$stream_type = "indexed";
	$index_range = 64*1024;
} elsif( $format =~ m/^cr\d+rw$/ ) {
	$format =~ /^cr?(\d+)rw$/;
	$element_size = $1;
	$stream_rw[1] = "WRITE";
	$stream_type = "indexed";
	$index_range = 64*1024;
} else {
	print "illegal format\n";
	exit;
}

print "element size: $element_size" . "B\n";

if( $stream_type eq "stride" ) {
	print "stride: $stride\n";
} elsif( $stream_type eq "stride_consec" ) {
	print "stride: $stride\n";
	print "consec num: $consec_num" . " elements\n";
} elsif( $stream_type eq "indexed" ) {
	print "index range: $index_range" . " elements\n";
} elsif( $stream_type eq "indexed_consec" ) {
	print "consec num: $consec_num" . " elements\n";
	print "index range: $index_range" . " elements\n";
} else {
	die;
}
print "stream interleaving: $stream_rw[0]: $stream_rw[1]\n";
print "number of streams: $num_stream\n";
print "stream length: $stream_length\n";
print "DRAM access granularity: $access_granularity\n";
print "---------------------------------------------------------------\n";

@base_addr = ( (0x00002aaae910000000), (0x00002aaae930000000), (0x00002aaae950000000), (0x00002aaae970000000), 
               (0x00002aaae990000000), (0x00002aaae9a0000000), (0x00002aaae9c0000000), (0x00002aaae9e0000000) );
@addr = @base_addr;



for( $j=0; $j<$num_stream; $j++ ) {
	$addr[$j] -= $access_granularity;
}

$access_granularity_bits = 3;
if( $access_granularity == 8 ) {
	$access_granularity_bits = 3;
} elsif( $access_granularity == 64 ) {
	$access_granularity_bits = 6;
} elsif( $access_granularity == 128 ) {
	$access_granularity_bits = 7;
} elsif( $access_granularity == 256 ) {
	$access_granularity_bits = 8;
}
else {
	die("access granularity: 8 or 64");
}


open (TRACE, ">", $trace_file_name) or die("Can't open trace fiel: $trace_file_name");
for( $i=0; $i<$stream_length; $i++ ) {
	for( $j=0; $j<$num_stream; $j++ ) {
		$operation = $stream_rw[$j % 2];
		if( $stream_type eq "stride" ) {
			$new_addr = $base_addr[$j] + $element_size*$i*$stride;
    } elsif( $stream_type eq "stride_consec" ) {
      if( ($i % $consec_num) == 0 ) {
			  $new_addr = $base_addr[$j] + $element_size*$i*$stride;
        $consec_base[$j] = $new_addr;
      } else {
        $new_addr = $consec_base[$j] + $element_size * ($i % $consec_num);
      }
		} elsif( $stream_type eq "indexed" ) {
			$new_addr = $base_addr[$j] + $element_size*ceil(rand($index_range-1));
		} elsif( $stream_type eq "indexed_consec" ) {
      if( ($i % $consec_num) == 0 ) {
        $new_addr = $base_addr[$j] + $element_size*ceil(rand($index_range-1));
        $consec_base[$j] = $new_addr;
      } else {
        $new_addr = $consec_base[$j] + $element_size * ($i % $consec_num);
      }
    }

		if( ($new_addr+$element_size <= ($addr[$j] + $access_granularity)) &&
				($new_addr >= $addr[$j]) ) {
			# skip this! -- emulating a single entry coalesce buffer!
		} else {
			$xxx = $new_addr;
#			printf "--- 0x%016x\n", $addr[$j];
			$xxx >>= $access_granularity_bits;
			$xxx <<= $access_granularity_bits;
			$addr[$j] = $xxx;
#			printf "~~~ 0x%016x\n", $addr[$j];
			printf TRACE ("0x%016x  $operation    $time    $access_granularity\n", $xxx);
		}
	}
	if( (($i+1) % $ag_width) == 0 ) { $time += $injection_interval; }
}
close(TRACE);

sub print_usage {
	print "usage: $0 format num_stream stream_length access_granularity trace_file_name\n";
	print "format: type1 -- AxB<suffix>\n";
	print "        type2 -- <prefix>A<suffix>\n";
	print "        type3 -- <prefix>A+C<suffix>\n";
	print "        A -- element size in B, 8B\n";
	print "        B -- stride in elements, 1: unitstride\n";
	print "        C -- number of consecutive elements following the random address, 1: random\n";
	print "        <prefix> -- r: random (4M elements), cr: constrained random (64k elements)\n";
	print "        <suffix> -- rd: read streams, rw: interleave read/write streams\n";
	print "        Examples) -- 8x1rd: 8B unit stride, r8rd: 8B random reads\n";
  print "                     r64+10rd: starting at random address, ten 64B consecutive reads\n";
	print "num_stream: number of streams (1..8) -- reqs from different streams are interleaved to the output trace\n";
	print "stream_length: number of elements in a stream (1..4096)\n";
	print "access_granularity: memory access granularity -- examples) 8B or 64B\n";
	print "trace_file_name: output trace file name\n";
	print "\n\n";
	exit;
}
